import os
import shutil
import stat
import time

from bygg.cmd.datastructures import ByggContext
from bygg.core.common_types import RunnerInstruction
from bygg.core.job import Job
from bygg.core.runner import get_job_count_limit
from bygg.output.job_output import output_job_logs
from bygg.output.output import (
    output_error,
    output_info,
    output_ok,
    output_plain,
    output_warning,
)
from bygg.output.status_display import failed_checks, output_check_results


def build(
    ctx: ByggContext,
    action: str,
    job_count: int | None,
    always_make: bool,
    check: bool,
) -> tuple[bool, set[str]]:
    """
    actions: The actions to build.

    job_count: The number of jobs to run simultaneously. None means to use the number of
    available cores.

    always_make: If True, all actions will be built, even if they are up to date.

    check: If True, apply various checks:

    * Check that the inputs and outputs of all actions will be checked for consistency.
      A job that runs later must not have files as output that are inputs to a job that
      runs earlier.

    Returns a tuple with build status and a set with found input files.
    """
    input_files: set[str] = set()
    try:
        max_workers = get_job_count_limit() if job_count is None else job_count

        t1 = time.time()
        output_info(f"Building action '{action}':")

        start_count = 0
        runner_instruction: RunnerInstruction | None = "restart_build"

        while runner_instruction == "restart_build":
            start_count += 1
            if start_count > 17:
                output_error("Too many restarts. Aborting.")
                return (False, input_files)

            ctx.scheduler.start_run(
                action,
                always_make=always_make,
                check=check,
            )

            # Collect for --watch. Needs to be done here when the graph is built up and
            # before build has started.
            for job_name in ctx.scheduler.job_graph.get_all_jobs():
                build_action = ctx.scheduler.build_actions.get(job_name)
                if build_action:
                    input_files.update(build_action.inputs)

            exit_reasons = ctx.runner.start(max_workers)
            ctx.scheduler.shutdown()
            runner_instruction = process_exit_reasons(exit_reasons)

            if runner_instruction is None:
                output_ok(f"Action '{action}' completed in {time.time() - t1:.2f} s.")
            elif runner_instruction == "restart_build":
                output_info("Restarting build.")
                pass
            else:
                output_error(
                    f"Action '{action}' failed after {time.time() - t1:.2f} s."
                )
                output_job_logs(ctx.runner.failed_jobs)
                return (False, input_files)

    except KeyboardInterrupt:
        output_plain("")
        output_warning("Build was interrupted by user.")
        return (False, input_files)
    except KeyError as e:
        output_error(f"Error: Action '{e}' not found.")
        return (False, input_files)
    finally:
        ctx.scheduler.shutdown()

    if check and failed_checks:
        return (output_check_results(), input_files)

    return (True, input_files)


def process_exit_reasons(exit_reasons: list[Job]) -> RunnerInstruction | None:
    if not exit_reasons:
        return None

    distilled_exit_reasons = {
        r.status.runner_instruction
        for r in exit_reasons
        if r.status and r.status.runner_instruction
    }

    if "restart_build" in distilled_exit_reasons:
        return "restart_build"

    return "exit_job_failed"


def clean(ctx: ByggContext, action: str):
    try:
        output_info(f"Cleaning action '{action}':")
        ctx.scheduler.prepare_run(action)
        for job_name in ctx.scheduler.job_graph.get_all_jobs():
            job = ctx.scheduler.build_actions.get(job_name, None)
            if job is None:
                continue
            for output in job.outputs:
                try:
                    s = os.stat(output)
                    if stat.S_ISREG(s.st_mode):
                        os.unlink(output)
                    elif stat.S_ISDIR(s.st_mode):
                        output_info(f"Removing directory: {output}")
                        shutil.rmtree(output)
                except FileNotFoundError:
                    pass
        ctx.scheduler.shutdown()
    except KeyboardInterrupt:
        output_plain("")
        output_warning("Build was interrupted by user.")
        return False
    except KeyError as e:
        output_error(f"Error: Action '{e}' not found.")
        return False
    finally:
        ctx.scheduler.shutdown()

    return True



import shutil

import nox

if shutil.which("uv"):
    nox.options.default_venv_backend = "uv"

nox.options.error_on_external_run = True
nox.options.stop_on_first_error

python_versions = ["3.11", "3.12", "3.13"]


# Not a test, but utility functionality for updating the help snapshots across all
# tested Python versions.
# Run with `nox -s update_help_snapshots`.
@nox.session(default=False, python=python_versions)
def update_help_snapshots(session):
    session.install("-r", "requirements.txt", "-r", "requirements-dev.txt")
    session.install(".")
    session.run("pytest", "-m", "help", "--snapshot-update")


@nox.session(python=python_versions)
def tests(session):
    session.install("-r", "requirements.txt", "-r", "requirements-dev.txt")
    session.install(".")
    session.run("pytest", "-n", "auto", "-vv", "-m", "not help")
    # The help test needs separate snapshots for each Python version, and the snapshot
    # package for pytest (syrupy) will report the snapshots for the Python version not
    # running currently as unused. Run the help test with the --snapshot-warn-unused
    # flag.
    session.run("pytest", "-vv", "-m", "help", "--snapshot-warn-unused")
    session.run("bygg", success_codes=[1], silent=True)
    session.run("bygg", "--help", silent=True)


@nox.session(python=python_versions)
def examples(session):
    session.install(".")

    with session.chdir("examples/only_python"):
        session.run("bygg", success_codes=[1])
        session.run("bygg", "hello")
        session.run("bygg", "hello", "--tree")
        session.run("bygg", "--clean", success_codes=[1])
        session.run("bygg", "hello", "--clean")

    with session.chdir("examples/parametric"):
        session.run("bygg")
        session.run("bygg", "--tree")
        session.run("bygg", "--clean")

    with session.chdir("examples/taskrunner"):
        session.run("bygg")
        session.run("bygg", "--tree")
        session.run("bygg", "--clean")

    with session.chdir("examples/trivial"):
        session.run("bygg")
        session.run("bygg", "transform")
        session.run("bygg", "--tree")
        session.run("bygg", "--clean")

    with session.chdir("examples/environments"):
        session.run("bygg")
        session.run("bygg", "clean_environments")

        session.run("bygg", "default_action")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action1")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action2")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action1", "default_action", "action2")
        session.run("bygg", "action1", "default_action", "action2", "--tree")
        session.run("bygg", "clean_environments")

    with session.chdir("examples/checks"):
        session.run("bygg")
        session.run("bygg", "--check")
        session.run("bygg", "--tree")
        session.run("bygg", "all_checks", "--check")
        session.run("bygg", "all_checks", "--tree")



import os
from pathlib import Path
import shutil
import subprocess
import sys
import textwrap
from typing import TYPE_CHECKING

from bygg.cmd.configuration import (
    DEFAULT_ENVIRONMENT_NAME,
    PYTHON_INPUTFILE,
    ActionItem,
    Byggfile,
    Environment,
)
from bygg.cmd.datastructures import ByggContext
from bygg.core.action import Action
from bygg.core.digest import calculate_string_digest
from bygg.logutils import logger
from bygg.output.output import output_error, output_info, output_plain
from bygg.util import create_shell_command


def calculate_environment_hash(environment: Environment) -> str:
    requirements: list[str] = []
    requirements.append(environment.shell)
    if environment.inputs:
        for input in environment.inputs:
            with open(input, "r") as f:
                requirements += f.readlines()
    return calculate_string_digest(" ".join(requirements))


def should_restart_with(environment: Environment) -> str | None:
    """Returns the path to the bygg file to restart with if it's not the one in the
    already loaded environment; in that case returns None."""
    venv_bin_path = Path(environment.venv_directory) / "bin"

    # # TODO: we could check here if the new venv is using the same Python binary that
    # # we're already running under, and if so just activate the venv.

    is_relative = Path(sys.executable).is_relative_to(venv_bin_path.resolve())
    if not is_relative:
        return str((venv_bin_path / "bygg"))
    return None


def setup_environment(ctx: ByggContext, environment: Environment):
    venv_path = Path(
        environment.venv_directory if environment.venv_directory else ".venv"
    )

    environment_hash = calculate_environment_hash(environment)
    environment_hash_file = venv_path / "bygg_environment_hash.txt"
    if environment_hash_file.exists():
        with open(environment_hash_file, "r") as f:
            if f.read() == environment_hash:
                return True
        environment_hash_file.unlink()

    # TODO should we remove the whole venv or just trust pip to do the right thing? For
    # now, remove the venv.

    environment_name = f' "{environment.name}" ' if environment.name else " "

    if venv_path.exists():
        output_info(f"Replacing venv{environment_name}at {venv_path}")
        shutil.rmtree(venv_path)

    output_info(f"Setting up environment{environment_name}in {venv_path}")
    process = subprocess.run(
        environment.shell,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )

    if process.returncode != 0:
        output_error("Error while creating virtual environment:")
        output_plain(process.stdout)
        sys.exit(1)

    if ctx.configuration.settings.verbose:
        output_plain(process.stdout)
        output_info("End environment setup output")

    with open(environment_hash_file, "w") as f:
        f.write(environment_hash)

    return True


def remove_environments(configuration: Byggfile) -> list[str]:
    removed_environments: list[str] = []
    for name, env in configuration.environments.items():
        venv_path = Path(env.venv_directory)
        if venv_path.exists():
            output_info(f"Removing venv {name} at {venv_path}")
            shutil.rmtree(venv_path)
            removed_environments.append(name)
    return removed_environments


def load_environment(ctx: ByggContext, environment_name: str):
    environment = ctx.configuration.environments.get(environment_name, None)

    # Now set up the actions for the current environment:
    register_actions_from_configuration(ctx.configuration, environment_name)

    # Evaluate the Python build file:

    # If the default Python byggfile is used in an environment, don't load it by default
    if environment_name == DEFAULT_ENVIRONMENT_NAME and PYTHON_INPUTFILE in {
        bf.byggfile for bf in ctx.configuration.environments.values()
    }:
        return None

    python_build_file = environment.byggfile if environment else PYTHON_INPUTFILE
    if python_build_file:
        load_python_build_file(python_build_file, environment_name)
    return None


def register_actions_from_configuration(
    configuration: Byggfile, is_restarted_with_env: str | None
):
    logger.info(
        "Registering actions from configuration for '%s'", is_restarted_with_env
    )
    for action_name, action in configuration.actions.items():
        if TYPE_CHECKING:
            assert not isinstance(action, str)

        if is_restarted_with_env and action.environment != is_restarted_with_env:
            logger.info(
                "Skipping action '%s' for environment '%s', it belongs to '%s'",
                action_name,
                is_restarted_with_env,
                action.environment,
            )
            continue

        logger.info("Registering action '%s'", action_name)
        logger.debug("Action: %s", action)
        shell_command = (
            create_shell_command(action.shell, action.message) if action.shell else None
        )
        Action(
            action_name,
            # Use the shell command as fallback for the description
            description=action.description or f"`{action.shell}`",
            is_entrypoint=bool(action.is_entrypoint),
            inputs=action.inputs,
            outputs=action.outputs,
            dependencies=action.dependencies,
            command=shell_command,
            environment=action.environment,
        )


def load_python_build_file(build_file: str, environment_name: str):
    # modify load path to make the current directory importable
    preamble = """\
        import os
        import sys
        sys.path.insert(0, str(os.path.realpath('.')))

        """

    if os.path.isfile(build_file):
        with open(build_file, "r") as f:
            Action._current_environment = environment_name
            logger.info(
                "Loading Python file %s for environment %s",
                build_file,
                environment_name,
            )
            exec(textwrap.dedent(preamble) + f.read(), globals())
            logger.info(
                "Back from loading Python file %s for environment %s",
                build_file,
                environment_name,
            )
            Action._current_environment = None


def get_environment_for_action(ctx: ByggContext, action_name: str) -> str | None:
    if isinstance(action := ctx.configuration.actions.get(action_name), ActionItem):
        return action.environment
    return None



import os
import sys

import pytest

from bygg.system_helpers import (
    ExitCode,
    call,
    change_dir,
    subprocess_tty,
    subprocess_tty_print,
)


def test_ExitCode():
    # Behaviour as a bool (inverted):
    assert ExitCode(0)
    assert not ExitCode(1)
    assert bool(ExitCode(0)) is True
    assert bool(ExitCode(1)) is False
    # Behaviour as an int:
    assert ExitCode(0) == 0
    assert ExitCode(0) != 1
    assert ExitCode(1) == 1
    assert ExitCode(1) != 0
    assert ExitCode(0) + 3 == 3
    assert ExitCode(1) + 3 == 4
    assert 3 + ExitCode(0) == 3
    assert 3 + ExitCode(1) == 4
    assert ExitCode(0) + ExitCode(0) == 0
    assert ExitCode(3) + ExitCode(2) == 5
    assert ExitCode(0) - ExitCode(0) == 0
    assert ExitCode(3) - ExitCode(2) == 1
    assert ExitCode(0) * ExitCode(0) == 0
    assert ExitCode(3) * ExitCode(2) == 6
    assert ExitCode(3) / ExitCode(2) == 1.5
    assert ExitCode(8) / ExitCode(-4) == -2

    with pytest.raises(ZeroDivisionError):
        assert ExitCode(0) / ExitCode(0) == 0


shell_command = "echo 'foo\nbar\nbaz'"


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="This test is problematic on macOS. See #337. ",
)
def test_call(capsys, snapshot):
    call(shell_command)
    captured = capsys.readouterr()
    assert captured.out == snapshot


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="This test is problematic on macOS. See #337. ",
)
def test_subprocess_tty_print(capsys, snapshot):
    subprocess_tty_print(shell_command.split(" "))
    captured = capsys.readouterr()
    assert captured.out == snapshot


def test_subprocess_tty():
    words = shell_command.split(" ")[-1].split("\n")
    for line, word in zip(subprocess_tty(shell_command.split(" ")), words):
        assert line.rstrip() == word


def test_change_dir():
    start_dir = os.getcwd()
    with change_dir("/"):
        assert os.getcwd() == "/"
    assert os.getcwd() == start_dir

    with change_dir("examples"):
        assert os.getcwd() == os.path.join(start_dir, "examples")
        with change_dir("checks"):
            assert os.getcwd() == os.path.join(start_dir, "examples/checks")
        assert os.getcwd() == os.path.join(start_dir, "examples")
    assert os.getcwd() == start_dir

    with change_dir("examples/checks"):
        assert os.getcwd() == os.path.join(start_dir, "examples/checks")
    assert os.getcwd() == start_dir

    with change_dir("examples/trivial"):
        assert os.getcwd() == os.path.join(start_dir, "examples/trivial")
    assert os.getcwd() == start_dir

    with change_dir(None):
        assert os.getcwd() == start_dir
    assert os.getcwd() == start_dir



import os
from pathlib import Path
import time

import pytest

from bygg.core.digest import (
    calculate_dependency_digest,
    calculate_file_digest,
    file_digest_memo,
)

pytestmark = pytest.mark.digest


def test_calculate_file_digest_existing(tmp_path):
    filenames = ["file1", "file2", "file3"]

    for filename in filenames:
        file = tmp_path / filename
        file.write_text(f"content for {filename}")

    digests1 = set(
        calculate_file_digest(str(tmp_path / filename)) for filename in filenames
    )
    assert all(digests1)

    digests2 = set(
        calculate_file_digest(str(tmp_path / filename)) for filename in filenames
    )

    assert digests1 == digests2


def test_calculate_file_digest_non_existing(tmp_path):
    filenames = ["file1", "file2", "file3"]

    digests1 = set(
        calculate_file_digest(str(tmp_path / filename)) for filename in filenames
    )
    assert map(lambda x: x is None, digests1)

    digests2 = set(
        calculate_file_digest(str(tmp_path / filename)) for filename in filenames
    )

    assert map(lambda x: x is None, digests2)


def test_calculate_directory_digest_existing(tmp_path):
    directory = tmp_path / "digest_directory"
    directory.mkdir()

    digest = calculate_file_digest(str(directory))
    assert digest == "directory"


def test_calculate_directory_digest_non_existing(tmp_path):
    directory = tmp_path / "digest_directory"

    digest = calculate_file_digest(str(directory))
    assert digest is None


def test_calculate_dependency_digest_existing(tmp_path):
    files = set(tmp_path / filename for filename in ["file1", "file2", "file3"])

    for file in files:
        file.write_text(f"content for {str(file)}")

    digests1, was_missing1 = calculate_dependency_digest(
        set(str(file) for file in files)
    )

    assert digests1
    assert not was_missing1

    digests2, was_missing2 = calculate_dependency_digest(
        set(str(file) for file in files)
    )

    assert digests1 == digests2
    assert not was_missing2


def test_calculate_dependency_digest_non_existing(tmp_path):
    files = list(tmp_path / filename for filename in ["file1", "file2", "file3"])

    for file in files:
        file.write_text(f"content for {str(file)}")

    digests1, was_missing1 = calculate_dependency_digest(
        set(str(file) for file in files)
    )

    assert digests1
    assert not was_missing1

    filenames = set(str(file) for file in files)
    unlinked_file = files.pop()
    unlinked_file.unlink()

    digests2, was_missing2 = calculate_dependency_digest(filenames)

    assert len(digests2)
    assert digests2 != digests1
    assert was_missing2


# Time to sleep after modifying file content that is different but has the same size.
# Simulates that it's unlikely that different content with the same size is written at
# the same time as the previous content that we digested.
SETTLE_TIME = 1 / 1e2


def test_file_digest_memo(tmp_path, mocker):
    import bygg.core.digest

    digest_spy = mocker.spy(bygg.core.digest, "file_digest")

    file = tmp_path / "file"
    filename = str(file)

    # Initial run, should give cache miss.
    expected_digest_calls = 1
    file.write_text(f"content for {filename}")
    st = os.stat(filename)
    digest1 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest1
    assert digest_spy.call_count == expected_digest_calls

    # No content or stat change. Should give cache hit.
    digest2 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest2 == digest1
    assert digest_spy.call_count == expected_digest_calls

    # Write the same content again. Should give cache miss.
    expected_digest_calls += 1
    time.sleep(SETTLE_TIME)
    file.write_text(f"content for {filename}")
    st = os.stat(filename)
    digest3 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest3 == digest1
    assert digest_spy.call_count == expected_digest_calls

    # Again, no content or stat change. Should give cache hit.
    digest4 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest4 == digest1
    assert digest_spy.call_count == expected_digest_calls

    # Write new content but with same size. Should give cache miss.
    expected_digest_calls += 1
    time.sleep(SETTLE_TIME)
    file.write_text(f"content_for_{filename}")
    st = os.stat(filename)
    digest5 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest5 != digest1
    assert digest_spy.call_count == expected_digest_calls

    # Write new content with different size. Should give cache miss.
    expected_digest_calls += 1
    file.write_text(f"more content for {filename}")
    st = os.stat(filename)
    digest6 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest6 != digest1
    assert digest_spy.call_count == expected_digest_calls

    # Again, no content or stat change. Should give cache hit.
    digest7 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest7 == digest6
    assert digest_spy.call_count == expected_digest_calls

    # No content or stat change. Should give cache hit.
    digest8 = file_digest_memo(filename, st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest8 == digest6
    assert digest_spy.call_count == expected_digest_calls


def test_calculate_dependency_digest(tmp_path: Path, mocker):
    file = tmp_path / "file"
    filename = str(file)
    file_set = set([filename])

    # Initial run
    file.write_text(f"content for {filename}")
    digests1, was_missing1 = calculate_dependency_digest(file_set)

    assert digests1
    assert not was_missing1

    # No content change
    time.sleep(SETTLE_TIME)
    file.write_text(f"content for {filename}")
    digests2, was_missing2 = calculate_dependency_digest(file_set)

    assert digests2 == digests1
    assert not was_missing2

    # Content change but no size change
    time.sleep(SETTLE_TIME)
    file.write_text(f"content_for_{filename}")
    digests3, was_missing3 = calculate_dependency_digest(file_set)

    assert digests3 != digests1
    assert not was_missing3

    # Content change with size change
    file.write_text(f"more content for {filename}")
    digests4, was_missing4 = calculate_dependency_digest(file_set)

    assert digests4 != digests3
    assert not was_missing4


def test_file_digest_memo_symlinks(tmp_path, mocker):
    import bygg.core.digest

    digest_spy = mocker.spy(bygg.core.digest, "file_digest")

    file = tmp_path / "real.txt"
    file.write_text(f"real file content for {str(file)}")

    link = tmp_path / "link.txt"
    os.symlink(file, link)

    st = os.stat(str(link))
    digest1 = file_digest_memo(str(file), st.st_ctime_ns, st.st_mtime_ns, st.st_size)
    digest2 = file_digest_memo(str(link), st.st_ctime_ns, st.st_mtime_ns, st.st_size)

    assert digest1 == digest2
    assert digest_spy.call_count == 2


def test_calculate_dependency_digest_symlinks(tmp_path: Path, mocker):
    import bygg.core.digest

    digest_spy = mocker.spy(bygg.core.digest, "file_digest")

    file = tmp_path / "file"
    file.write_text(f"content for {str(file)}")

    link = tmp_path / "link"
    link.symlink_to(file)

    digests1, was_missing1 = calculate_dependency_digest(set([str(link)]))

    assert digests1
    assert not was_missing1
    assert digest_spy.call_count == 1

    file.write_text(f"different content for {str(file)}")

    digests2, was_missing2 = calculate_dependency_digest(set([str(link)]))

    assert digests1 != digests2
    assert not was_missing2
    assert digest_spy.call_count == 2



"""
System-near helper functions. Only import things from the standard library here; this
way these helpers can be used also for utility scripts.
"""

import contextlib
import errno
import os
import pty
import shlex
import signal
import subprocess
from typing import Generator


class ProcessGenerator:
    """
    Class that wraps a process return value and a generator that yields the output of a
    process line by line.

    Attributes
    ----------
    returncode : int
        Process return code.
    generator : Generator
        Generator that yields the output of the process line by line.

    Yields
    ------
    str
        Output of the process.
    """

    returncode: int
    """Process return code."""

    generator: Generator
    """Generator that yields the output of the process line by line."""

    def __init__(self):
        self.returncode = -1

    def __iter__(self):
        self.value = yield from self.generator


def subprocess_tty(cmd, encoding="utf-8", timeout=10, **kwargs):
    """
    Wrapper around subprocess.Popen that sets up the process as if it were running in a
    TTY and returns a generator that yields the stdout of the process line by line.

    Parameters
    ----------
    cmd :
        Passed directly to subprocess.Popen; see its documentation for details.
    encoding : str, optional
        Encoding to use when reading stdout, by default "utf-8".
    timeout : int, optional
        Timeout to use when polling the child process to see if it has exited, by
        default 10 seconds.

    Returns
    -------
    ProcessGenerator
        Generator that yields the stdout of the child process line by line. After the
        generator is exhausted, the returncode attribute of the generator will be set to
        the exit code of the child process.

    Yields
    ------
    str
        Lines of stdout from the child process.

    Raises
    ------
    subprocess.TimeoutExpired
        Re-raised if the child process did not exit within the set timeout after having
        received both terminate and kill signals.

    Can also raise other exceptions from subprocess.Popen.
    """

    generator = ProcessGenerator()

    def subprocess_iterator():
        # From https://stackoverflow.com/a/77387332
        m, s = pty.openpty()
        p = subprocess.Popen(cmd, stdout=s, stderr=s, **kwargs)
        os.close(s)

        try:
            for line in open(m, encoding=encoding):
                if not line:  # EOF
                    break
                yield line
        except OSError as e:
            if e.errno != errno.EIO:  # EIO also means EOF
                raise
        finally:
            if p.poll() is None:
                p.send_signal(signal.SIGINT)
                try:
                    p.wait(timeout)
                except subprocess.TimeoutExpired:
                    p.terminate()
                    try:
                        p.wait(timeout)
                    except subprocess.TimeoutExpired:
                        p.kill()
                        raise
            p.wait()
            generator.returncode = p.returncode

    generator.generator = subprocess_iterator()
    return generator


def subprocess_tty_print(cmd, encoding="utf-8", timeout=10, **kwargs):
    """
    Thin wrapper around subprocess_tty that prints its output line by line. See
    subprocess_tty for details.


    Parameters
    ----------
    See subprocess_tty.

    Returns
    -------
    int
        The status code from the subprocess.
    """
    proc = subprocess_tty(cmd, encoding, timeout, **kwargs)
    for line in proc:
        print(line.rstrip())
    return proc.returncode


class ExitCode(int):
    """
    An int with a customised __bool__ method that considers 0 to be truthy. This is
    useful for working with subprocess return codes.
    """

    def __bool__(self):
        return self == 0


def call(cmd, encoding="utf-8", timeout=10, **kwargs) -> ExitCode:
    """
    Convenience wrapper around subprocess_tty_print; the command to be executed can be
    given as string that will be split instead of as an array. Output will be printed
    line by line.

    Parameters
    ----------
    cmd : str
        The command to be executed. Will be run through shlex.split().

    See subprocess_tty_print and subprocess_tty for the other parameters.

    Returns
    -------
    ExitCode
        The exit code from the subprocess. ExitCode is a subclass of int that can be
        used as a boolean; subprocess success (0) is considered truthy.

    Raises
    ------
    See subprocess_tty_print and subprocess_tty.
    """

    return ExitCode(subprocess_tty_print(shlex.split(cmd), encoding, timeout, **kwargs))


@contextlib.contextmanager
def change_dir(dir: str | os.PathLike | None):
    old_dir = os.getcwd()
    try:
        if dir is not None:
            os.chdir(dir)
        yield
    finally:
        os.chdir(old_dir)



from bygg.core.action import Action, ActionContext, action
from bygg.core.common_types import CommandStatus


def hello(ctx: ActionContext):
    """An action that says 'Hello'."""
    print("Hello")
    return CommandStatus(0, "And goodbye.", None)


Action("hello", command=hello, is_entrypoint=True)


@action("hi", is_entrypoint=True)
def hi(ctx: ActionContext):
    """An action that says 'Hi'."""
    print("Hi")
    return CommandStatus(0, "And goodbye.", None)



import os
import random
import subprocess
import time

from bygg.core.action import Action, ActionContext, CommandStatus

# Modify these constants to change aspects of the test build:

# Directory for output files
LEVELLED_DIR_PATH = "t"

# How many levels to build up
TEST_LEVELS = 3

# How many files per level
TEST_FILES_PER_LEVEL = 50

# Do something more than just sort lines in files. This will calculate prime numbers.
GENERATE_LOAD = False

# Sleep in each job. Reduces throughput; useful to check logging and output. Don't sleep
# in CI.
ADD_SLEEP = False if "CI" in os.environ else True

# Execute touch <outfile> in a shell instead of sorting the test files in Python. This
# will cause work to be done by the process pool, but since the work is trivial, it
# won't hardly be noticeable.
USE_SHELL_COMMAND = False


def calculate_first_prime_numbers(count: int) -> list[int]:
    """Calculate the first n prime numbers."""
    primes: list[int] = []
    i = 2
    while len(primes) < count:
        if all(i % prime != 0 for prime in primes):
            primes.append(i)
        i += 1
    return primes


# Action function for sorting the input files.
def sort_lines(ctx: ActionContext):
    if not ctx.inputs:
        return CommandStatus(1, "No inputs.", None)

    changed_files: set[str] = set()
    output_lines: list[str] = []

    for input in ctx.inputs:
        output = input[:-3]
        output_lines.append(f"{input} -> {output}")
        with open(input, "r") as f:
            lines = f.readlines()
        lines.sort()
        with open(output, "w") as f:
            f.writelines(lines)
        changed_files.add(output)
    sleep_time = random.random() if ADD_SLEEP else 0
    output_lines.append(f"Slept for {sleep_time} seconds")
    time.sleep(sleep_time)

    if GENERATE_LOAD:
        calculate_first_prime_numbers(5000)
    return CommandStatus(0, "Sorted lines.", "\n".join(output_lines))


def shell_executor(command: str) -> tuple[int, str]:
    process = subprocess.run(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )
    return process.returncode, process.stdout


def call_shell_command(ctx: ActionContext):
    sleep_time = random.random() if ADD_SLEEP else 0
    time.sleep(sleep_time)
    output_lines: list[str] = []
    rc_list: list[int] = []
    for output_file in ctx.outputs:
        rc, output = shell_executor(f"touch {output_file}")
        rc_list.append(rc)
        output_lines.append(output)
    return CommandStatus(
        max(rc_list),
        f"Touched output file in {sleep_time:.2f} s",
        "\n".join(output_lines),
    )


def generate_create_action(path: str):
    def action(ctx: ActionContext):
        if not os.path.isdir(LEVELLED_DIR_PATH):
            os.makedirs(LEVELLED_DIR_PATH, exist_ok=True)
        with open(path, "w") as f:
            f.write("foo\nbar\nbaz\n")
        return CommandStatus(0, "Created test file.", None)

    return action


def declare_test_files_actions():
    generated_files: list[str] = []
    generated_output_actions: list[Action] = []

    # Create a set of files with multiple dependency levels. In a real-world scenario,
    # these files would most likely already be present in the filesystem as source files
    # and listed out by a glob.

    def get_filename(level: int, file: int):
        return os.path.join(LEVELLED_DIR_PATH, f"l{level}f{file}.txt.in")

    for level in range(TEST_LEVELS):
        for file in range(TEST_FILES_PER_LEVEL):
            path = get_filename(level, file)
            generated_files += [path]

            Action(
                name=path,
                message=f"Creating test file {path}",
                inputs=None,
                outputs=[path],
                command=generate_create_action(path),
                scheduling_type="in-process",
            )

            # Add an action for an output file.
            output_file = path[:-3]
            name = output_file
            dependencies = [path]
            if level < TEST_LEVELS - 1:
                dependencies += [get_filename(level + 1, file)[:-3]]
            else:
                dependencies += []

            generated_output_actions += [
                Action(
                    name=name,
                    message=f"Sorting test file {path}",
                    inputs=[path],
                    outputs=[output_file],
                    dependencies=dependencies,
                    command=call_shell_command if USE_SHELL_COMMAND else sort_lines,
                )
            ]

    # Create the dummy input files
    Action(
        name="setup_test_files",
        outputs=generated_files,
        dependencies=[f for f in generated_files],
        message="Setting up test files.",
    )

    # Perform a "build" job (execute all the sort actions)
    Action(
        name="sort_test_files",
        dependencies=[
            a.name
            for a in generated_output_actions
            if a.name.startswith(f"{LEVELLED_DIR_PATH}/l0")
        ],
        outputs=["t"],
        message="Sorting test files.",
    )


declare_test_files_actions()


if __name__ == "__main__":
    declare_test_files_actions()
